-------------------------------------------------------------------------------
-- _    _  ____  ____  ____  ____  __  __  _  _  ____  __    ____ 
--( \/\/ )(_  _)(  _ \( ___)(  _ \(  )(  )( \( )(  _ \(  )  ( ___)
-- )    (  _)(_  )   / )__)  ) _ < )(__)(  )  (  )(_) ))(__  )__) 
--(__/\__)(____)(_)\_)(____)(____/(______)(_)\_)(____/(____)(____)
--
-- By Alejo "Bluesummers" Grigera (MrBluesummers@MrBluesummers.com)
-- Based on WireJumble by Neil Blevins
-- Created On: 03/15/09
-- Tested using Max 2009 sp1
-- Best viewed in mono-spaced font.
-- Please do not re-distribute this script, just visit my site.  :)
-------------------------------------------------------------------------------

-- Macroscript Definition
--		--		--		--		--
macroscript WireBundle category:"Bluesummers" tooltip:"Wire Bundle"
(
	try
	(
		-- Variables
		-- TO DO: ORGANIZE THESE
		local isRenderable = true
		local isViewable = true
		local Sides = 12
		local Steps = 6
		local NumberofSplines = 12
		local isAdaptive = false;
		local isOptimized = true;
		local RenderVisPercent = 100
		local ViewportVisPercent = 100
		local PathingMode = "spline"
		local InterpretationMode = 1
		local WireThickness = 1;
		local ThicknessVariance = 25;
		local Nodelist = #()
		local TargetSpline;
		local LastBuild = #()
		
		-- Function Definition	-----------------------------------
		-----------------------------------------------------------
		
		-- DEFINE SELECTOR VALIDITY
		fn PickButtonFilter obj = superClassOf obj == GeometryClass
		fn SplineButtonFilter spline = superClassOf spline == Shape
		
		-- SORT TWO OBJECTS BY NAME
		fn BulkInsertQSort v1 v2 =
		(
			w1 = v1.name
			w2 = v2.name
			
			stricmp w1 w2
		)
		
		-- OBJECT FLOW PROCESSING
		fn GetObjectPoint Obj =
		(
			print ("Object Interp: " + InterpretationMode as string)
			case InterpretationMode of
			(
				1: -- BOUNDING BOX METHOD
				(
					x = (random Obj.max.x Obj.min.x)
					y = (random Obj.max.y Obj.min.y)
					z = (random Obj.max.z Obj.min.z)
				)
				2: -- RANDOM VERTEX METHOD
				(
					-- Get a mesh
					MeshOBJ = snapshot Obj
					
					-- Get a random vertex
					vert_index = random 1 MeshOBJ.numverts
					x = (getVert MeshOBJ vert_index).x
					y = (getVert MeshOBJ vert_index).y
					z = (getVert MeshOBJ vert_index).z
					delete MeshOBJ
				)
				3: -- BOUNDING SPHERE METHOD
				(
					-- Someday.
				)
				4: -- TRUE VOLUME (RAYCAST METHOD, SLOW)
				(
					-- Someday.
				)
				default: ()
			)
			return ([x,y,z] as point3)
		)
		
		-- SPLINE FLOW PROCESSING
		fn GetSplinePoint Target Obj=
		(
			Thickness = Obj.Render_Thickness
			case InterpretationMode of
			(
				1: -- STAY WITHIN VOLUME
				(
					x = Target.x + Random (-Thickness/2.0) (Thickness/2.0)
					y = Target.y + Random (-Thickness/2.0) (Thickness/2.0)
					z = Target.z + Random (-Thickness/2.0) (Thickness/2.0)
				)
				2: -- RUN ALONG THE SURFACE (WITH TWIST)
				(
					-- Get a temporary Mesh
					TempObj = copy Target
					TempObj.render_displayRenderMesh = true
					TempObj.render_sides = NumberofSplines
					TempMesh = snapshot Target
					delete TempObj
					
					-- Add splines along the length of the mesh.
					-- Someday
				)
			)
			return ([x,y,z] as point3)
		)
		
		fn GetSplinePointHandles ThisShape ThisLine ThisPoint =
		(
			ReturnValues = #()
			ReturnValues[1] = GetInVec ThisShape ThisLine ThisPoint
			ReturnValues[2] = GetOutVec ThisShape ThisLine ThisPoint
			return ReturnValues
		)
		
		-- GET NODE LIST AS NAME LIST
		fn GetNodesAsStrings List =	
		(
			StringsList = #()
			for i = 1 to List.Count do StringsList = Append StringsList List[i].Name
			return StringsList
		)
		
		-- 
		fn BuildToFollowObjects ThisSpline Index =
		(
			for Obj in Nodelist do
			(
				TargetPoint = GetObjectPoint Obj
				AddKnot ThisSpline Index #Smooth #Curve TargetPoint
			)
		)
		
		fn BuildToFollowSpline ThisSpline Index =
		(
			PathSpline = ConvertToSplineShape TargetSpline
			print PathSpline
			
			for i = 1 to NumKnots PathSpline do
			(
				print i
				TargetPoint = GetSplinePoint (GetKnotPoint PathSpline 1 i) ThisSpline
				TargetPointType = GetKnotType PathSpline 1 i
				TargetHandles = GetSplinePointHandles PathSpline 1 i
				print TargetPoint
				AddKnot ThisSpline Index TargetPointType #Curve TargetPoint TargetHandles[1] TargetHandles[2]
			)		
		)

		-- MAIN
		fn Begin =
		(
			-- Validate Inputs
			print PathingMode
			print TargetSpline
			
			if PathingMode == "spline" and TargetSpline == undefined then
				messageBox "You must select a spline first."
			else if PathingMode == "objects" and Nodelist.count < 1 then
				messageBox "You must select more than one object to draw wires through."
			else
			(
				-- Initialize
				max select none
				LastBuild = #()
				WireColor = (color (random 0 255) (random 0 255) (random 0 255))
				
				for i = 1 to NumberofSplines do
				(
					NewSpline = line pos:[0, 0, 0]
					CurrentSplineIndex = AddNewSpline NewSpline
					
					-- The spline builder is dependent on path type.
					if PathingMode == "spline"
					then BuildToFollowSpline NewSpline CurrentSplineIndex
					else BuildToFollowObjects NewSpline CurrentSplineIndex
					
					-- Setup parameters on the new spline.
					NewSpline.Steps = Steps
					NewSpline.Optimize = isOptimized
					NewSpline.Adaptive = isAdaptive
					
					NewSpline.Render_Thickness = WireThickness + ((random (-ThicknessVariance) ThicknessVariance) * (WireThickness/100.0))
					
					NewSpline.Render_Renderable = isRenderable and ((i*100.0)/NumberOfSplines) < RenderVisPercent
					NewSpline.Render_DisplayRenderMesh  = isViewable and ((i*100.0)/NumberOfSplines) < ViewportVisPercent
					NewSpline.Render_MapCoords = true;
					NewSpline.Name = "WireBundle_" + (i as string)
					NewSpline.WireColor = WireColor
					append LastBuild NewSpline
				)
			)
		)		
		
		-- Rollout	
		rollout WireBundle "WireBundle Beta" width:250 height:452
		(
			spinner SidesSpinner "Sides:  " pos:[29,40] width:87 height:16 range:[0,256,12] type:#integer scale:0.2 fieldwidth:40
			spinner numWires "Quantity:  " pos:[16,64] width:100 height:16 range:[1,256,12] type:#integer scale:0.2 fieldwidth:40
			
			-- Interpolation
			spinner StepsSpinner "Steps:  " pos:[27,88] width:88 height:16 range:[0,256,6] type:#integer scale:0.2 fieldwidth:40
			checkbox adaptive "Adaptive" pos:[144,112] width:66 height:15 checked:false
			checkbox Optimized "Optimize" pos:[64,112] width:64 height:15 checked:true
				
			-- Visibility
			checkbox EnableInRenderer "Enable In Renderer" pos:[20,153] width:116 height:15 checked:true
			spinner RenderVisVar "%:  " pos:[155,153] width:77 height:16 range:[0,100,100] type:#integer scale:1 fieldwidth:40
			checkbox EnableInViewport "Enable In Viewport" pos:[20,173] width:113 height:15 checked:true
			spinner ViewportVisVar "%:  " pos:[155,173] width:77 height:16 range:[0,100,100] type:#integer scale:1 fieldwidth:40

			-- Extras
			radiobuttons PathingModeRadios "Pathing Method" pos:[15,220] width:81 height:46 labels:#("Use Spline", "Use Objects") columns:1 across:2
			listbox PathObjects "    Path Objects" pos:[96,224] width:145 height:9 enabled:false
			dropdownList ObjectInterp "" pos:[105,368] width:130 height:21 items:#("Random Pathfollow")
			pickbutton AddObject "Add..." pos:[13,296] width:44 height:21 enabled:false --filter:PickButtonFilter
			button AddSelObj "Sel" pos:[61,296] width:27 height:21 enabled:false
			button RemoveSelected "Remove" pos:[13,320] width:75 height:21 enabled:false
			button MoveUp "Move Up" pos:[13,345] width:75 height:21 enabled:false
			button MoveDown "Move Down" pos:[13,370] width:75 height:21 enabled:false

			-- Link
			HyperLink WebsiteLink "Written by Mr. Bluesummers, 2009" pos:[16,432] width:161 height:15 address:"http://www.MrBluesummers.com" across:2
			HyperLink HelpLink "Help" pos:[218,432] width:22 height:15 address:"http://www.mrbluesummers.com/874/video-tutorials/wire-bundle-monday-movie"
			
			-- Group Boxes
			GroupBox grp22 "Spline Parameters" pos:[5,5] width:240 height:196
			GroupBox grp24 "Quantity and Quality" pos:[11,20] width:229 height:108
			GroupBox grp25 "Visibility" pos:[11,133] width:229 height:63
			GroupBox grp26 "Bundle Pathing" pos:[5,205] width:241 height:190
			pickbutton PickSpline "Pick Spline" pos:[13,272] width:75 height:21 across:4 --filter:SplineButtonFilter
			button CreateButton "Create Bundle" pos:[152,401] width:88 height:25
			button CloseButton "Close" pos:[8,401] width:48 height:25
			button DeleteLastButton "Delete Last" pos:[64,400] width:80 height:25
			button ResetButton "Reset" pos:[190,222] width:50 height:15 enabled:false
			
			
			spinner ThicknessSpinner "Thickness:  " pos:[144,40] width:87 height:16 range:[0.001,999,1] type:#float scale:0.01 fieldwidth:40
			spinner ThicknessVar "Var %:  " pos:[145,64] width:87 height:16 range:[0,500,25] type:#integer scale:1 fieldwidth:40
			
			on AddSelObj pressed do 
			(
				SelectedNodes = $ as array
				(qsort SelectedNodes BulkInsertQSort)
				for i = 1 to SelectedNodes.Count do Append Nodelist SelectedNodes[i]
				PathObjects.Items = GetNodesAsStrings Nodelist
			)
			on ObjectInterp selected state 	do InterpretationMode = ObjectInterp.Selection
			on SidesSpinner changed value 	do Sides = value
			on StepsSpinner changed value 	do Steps = value
			on NumWires changed value 		do NumberofSplines = value
			on Adaptive changed value 		do (isAdaptive = Adaptive.checked; StepsSpinner.enabled = not value)
			on Optimized changed value 		do isOptimized = Optimized.checked
			on EnableInRenderer changed value do (isRenderable = EnableInRenderer.checked; RenderVisVar.enabled = value)
			on RenderVisVar changed value 	do RenderVisPercent = value
			on EnableInViewport changed value do (isViewable = EnableInViewport.checked; ViewportVisVar.enabled = value)
			on ViewportVisVar changed value do ViewportVisPercent = value
			on ThicknessSpinner changed value do WireThickness = value;
			on ThicknessVar changed value do ThicknessVariance = value;
			on ResetButton pressed do
			(
				Nodelist = #()
				PathObjects.Items = GetNodesAsStrings Nodelist
			)
			
			on PathingModeRadios changed state do
			(
				if PathingModeRadios.state == 1
				then PathingMode = "spline"
				else PathingMode = "objects"
					
				StateZero = PathingModeRAdios.state == 1
				
				PickSpline.enabled = StateZero
				AddObject.enabled = RemoveSelected.enabled = MoveUp.enabled =\
					MoveDown.enabled = PathObjects.Enabled = ResetButton.Enabled =\
					AddSelObj.enabled = not StateZero
				
				-->> Reset some UI Items
				NodeList = #()
				PickSpline.text = "Pick Spline"
				
				if StateZero
				then
				(
					ObjectInterp.Items = #("Random Pathfollow")--, "Orderly Pathfollow", "Twisted Pathfollow")
				)
				else
				(
					ObjectInterp.Items = #("Bounding Box", "Vertices")
				)
			)
			
			-->> OBJECT BOX HANDLERS
			on AddObject picked target do
				if target != undefined do
				(
					Nodelist = append Nodelist target
					PathObjects.Items = GetNodesAsStrings Nodelist
				)
				
			on PickSpline picked target do
				if target != undefined do
				(
					TargetSpline = target
					PickSpline.text = "[Spline Ready]"
				)
				
			on RemoveSelected pressed do
				if (PathObjects.Selection == 0)
				then MessageBox "You must select an object in the Path Objects box to remove!" \
						title:"Cannot Remove Nothing!"
				else 
				(
					NodeList = DeleteItem Nodelist PathObjects.Selection
					PathObjects.Items = GetNodesAsStrings Nodelist
				)
			
			on MoveUp pressed do
				if (PathObjects.Selection == 0)
				then MessageBox "You must select an object in the Path Objects box to move!" \
						title:"Cannot Move Nothing!"
				else if (PathObjects.Selection > 1) do
					(
						MovableObject = Nodelist[PathObjects.Selection]
						MovableObjectPos = PathObjects.Selection
						DeleteItem Nodelist PathObjects.Selection
						InsertItem MovableObject Nodelist (MovableObjectPos-1)
						PathObjects.Items = GetNodesAsStrings Nodelist
						PathObjects.Selection = PathObjects.Selection-1
					)
			on MoveDown pressed do
				if (PathObjects.Selection == 0)
				then MessageBox "You must select an object in the Path Objects box to move!" \
						title:"Cannot Move Nothing!"
				else if (PathObjects.Selection < PathObjects.Items.Count) do
					(
						MovableObject = Nodelist[PathObjects.Selection]
						MovableObjectPos = PathObjects.Selection
						DeleteItem Nodelist PathObjects.Selection
						InsertItem MovableObject Nodelist (MovableObjectPos+1)
						PathObjects.Items = GetNodesAsStrings Nodelist
						PathObjects.Selection = PathObjects.Selection+1
					)
			on CreateButton pressed do Begin()
			on CloseButton pressed do DestroyDialog WireBundle
			on DeleteLastButton pressed do try(for obj in LastBuild do delete obj) catch()
		)

		createdialog WireBundle lockheight:true lockwidth:true escapeEnable:true style:#(#style_toolwindow)
	) 
	catch
	(
		Messagebox "Bad things just happened.  Closing script"
		destroydialog WireBundle
	)
)

